/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.bridge.models;

import com.azure.storage.blob.models.BlobItem;
import org.apache.atlas.adls.bridge.ADLSDataTypes;
import org.apache.atlas.adls.bridge.ADLSHook;
import org.apache.atlas.adls.extractors.Constants;
import org.apache.atlas.adls.extractors.ADLSExtractorContext;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class AtlasADLSBlobEntity implements AtlasADLSEntity {
    private static final Logger LOG = LoggerFactory.getLogger(AtlasADLSBlobEntity.class);

    private static final String PROPERTY_LAST_MODIFICATION_TIME = "lastModifiedTime";
    private static final String PROPERTY_CREATION_TIME          = "creationTime";
    private static final String PROPERTY_ETAG                   = "eTag";
    private static final String PROPERTY_CONTENT_LENGTH         = "contentLength";
    private static final String PROPERTY_CONTENT_TYPE           = "contentType";
    private static final String PROPERTY_CONTENT_MD5            = "contentMd5";
    private static final String PROPERTY_CONTENT_ENCODING       = "contentEncoding";
    private static final String PROPERTY_CONTENT_DISPOSITION    = "contentDisposition";
    private static final String PROPERTY_CACHE_CONTROL          = "cacheControl";
    private static final String PROPERTY_BLOB_TYPE              = "type";
    private static final String PROPERTY_ACCESS_TIER            = "accessTier";
    private static final String PROPERTY_REMAINING_RETENTION    = "remainingRetentionDays";
    private static final String PROPERTY_PREFIX                 = "prefix";
    private static final String PROPERTY_ADDITIONAL             = "additionalProperties";

    private final ADLSExtractorContext extractorContext;
    private final String               containerAndStorageAccountName;
    private final BlobItem             blobItem;
    private final AtlasEntity          entity;
    private final String               entityType;

    public AtlasADLSBlobEntity(ADLSExtractorContext extractorContext, String containerAndStorageAccountName, BlobItem blobItem) {
        this(extractorContext, containerAndStorageAccountName, blobItem, ADLSDataTypes.ADLS_GEN2_BLOB.getEntityType());
    }

    public AtlasADLSBlobEntity(ADLSExtractorContext extractorContext, String containerAndStorageAccountName, BlobItem blobItem, String entityType) {
        if (StringUtils.isEmpty(containerAndStorageAccountName)) {
            String msg = "Illegal Azure Data Lake Storage container received.";

            throw new IllegalArgumentException(msg);
        }

        if (blobItem == null || StringUtils.isEmpty(blobItem.getName())) {
            String msg = "Illegal Azure Data Lake Storage Blob received.";

            throw new IllegalArgumentException(msg);
        }

        this.containerAndStorageAccountName = containerAndStorageAccountName;
        this.extractorContext               = extractorContext;
        this.blobItem                       = blobItem;
        this.entityType                     = entityType;
        this.entity                         = constructAtlasEntity();
    }

    public String getQualifiedName() {
        String fileSystemScheme            = extractorContext.getFileSystemScheme().toLowerCase();
        String containerAndStorageAcctName = containerAndStorageAccountName.toLowerCase();
        String prefixDelimiter             = extractorContext.getPrefixDelimiter();
        String metadataNamespace           = getMetadataNamespace();

        return fileSystemScheme + SCHEME_SEPARATOR + containerAndStorageAcctName +
                prefixDelimiter + getPrefix() + QNAME_SEP_METADATA_NAMESPACE + metadataNamespace;
    }

    private String getMetadataNamespace() {
        return ADLSHook.getInstance() != null ?
                ADLSHook.getInstance().getMetadataNamespace() : DEFAULT_METADATA_NAMESPACE;
    }

    @Override
    public String name() {
        String name = blobItem.getName();

        if (isDirectory(name)) {
            name += extractorContext.getPrefixDelimiter();
        }

        return name;
    }

    @Override
    public String getParentName() {
        return containerAndStorageAccountName;
    }

    @Override
    public AtlasEntity getEntity() {
        return entity;
    }

    @Override
    public AtlasEntity.AtlasEntitiesWithExtInfo getAtlasEntitiesWithExtInfo() {
        return new AtlasEntity.AtlasEntitiesWithExtInfo(this.getEntity());
    }

    protected AtlasEntity constructAtlasEntity() {
        AtlasEntity entity = new AtlasEntity(entityType);

        entity.setAttribute(PROPERTY_QUALIFIED_NAME, getQualifiedName());
        entity.setAttribute(PROPERTY_NAME, getEntityName(blobItem.getName()));
        entity.setAttribute(PROPERTY_PREFIX, getPrefix());

        if (blobItem.getProperties() != null) {
            entity.setAttribute(PROPERTY_LAST_MODIFICATION_TIME, getLastModifiedTime());
            entity.setAttribute(PROPERTY_CREATION_TIME, getCreationTime());
            entity.setAttribute(PROPERTY_ETAG, blobItem.getProperties().getETag());
            entity.setAttribute(PROPERTY_CONTENT_LENGTH, blobItem.getProperties().getContentLength());
            entity.setAttribute(PROPERTY_CONTENT_TYPE, blobItem.getProperties().getContentType());
            entity.setAttribute(PROPERTY_CONTENT_MD5, blobItem.getProperties().getContentMd5());
            entity.setAttribute(PROPERTY_CONTENT_ENCODING, blobItem.getProperties().getContentEncoding());
            entity.setAttribute(PROPERTY_CONTENT_DISPOSITION, blobItem.getProperties().getContentDisposition());
            entity.setAttribute(PROPERTY_CACHE_CONTROL, blobItem.getProperties().getCacheControl());
            entity.setAttribute(PROPERTY_BLOB_TYPE, blobItem.getProperties().getBlobType().toString());
            entity.setAttribute(PROPERTY_ACCESS_TIER, blobItem.getProperties().getAccessTier().toString());
            entity.setAttribute(PROPERTY_REMAINING_RETENTION, blobItem.getProperties().getRemainingRetentionDays());
            entity.setAttribute(PROPERTY_ADDITIONAL, blobItem.getMetadata());
        }

        if (LOG.isDebugEnabled()) {
            LOG.debug("Constructed Atlas entity with name {} and qualified name {}", getEntityName(blobItem.getName()), getQualifiedName());
        }

        return entity;
    }

    private boolean isDirectory(String name) {
        return MapUtils.isNotEmpty(blobItem.getMetadata())
                && blobItem.getMetadata().containsKey(Constants.IS_DIRECTORY_METADATA_KEY)
                && blobItem.getMetadata().get(Constants.IS_DIRECTORY_METADATA_KEY).equalsIgnoreCase(Boolean.TRUE.toString())
                && !doesObjectKeyEndsWithDirectorySeparator(name);
    }

    private Long getLastModifiedTime() {
        return blobItem.getProperties().getLastModified() != null ?
                blobItem.getProperties().getLastModified().toInstant().toEpochMilli() : null;
    }

    private Long getCreationTime() {
        return blobItem.getProperties().getCreationTime() != null ?
                blobItem.getProperties().getCreationTime().toInstant().toEpochMilli() : null;
    }

    private String getEntityName(String blobPath) {
        String ret = blobPath;

        if (StringUtils.isNotBlank(blobPath)) {
            int endPosWithoutLastDirectorySeparator = blobPath.length() - 1;

            if (doesObjectKeyEndsWithDirectorySeparator(blobPath)) {
                endPosWithoutLastDirectorySeparator--;
            }

            int nextDirectorySeparatorPositionFromLast = blobPath.lastIndexOf(extractorContext.getPrefixDelimiter(), endPosWithoutLastDirectorySeparator);

            ret = blobPath.substring(nextDirectorySeparatorPositionFromLast + 1, endPosWithoutLastDirectorySeparator + 1);
        }

        return ret;
    }

    private String getPrefix() {
        String prefix = blobItem.getName();

        if (doesObjectKeyEndsWithDirectorySeparator(prefix)) {
            prefix = StringUtils.chop(prefix);
        }

        return prefix;
    }

    private boolean doesObjectKeyEndsWithDirectorySeparator(String blobName) {
        return blobName.endsWith(extractorContext.getPrefixDelimiter());
    }
}