/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.adls.cli;

import com.google.common.annotations.VisibleForTesting;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.extractors.ExtractorTask;
import org.apache.atlas.adls.extractors.ExtractorTaskFactory;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;

public class ADLSExtractorMain {
    private static final Logger LOG = LoggerFactory.getLogger(ADLSExtractorMain.class);

    public static final int EXIT_CODE_SUCCESS = 0;
    public static final int EXIT_CODE_FAILED  = 1;
    public static final int EXIT_CODE_HELP    = 2;

    public static final String AZURE_EXTRACTOR_SCRIPT_NAME = "adls-extractor.sh";

    @VisibleForTesting
    static int exitCode = EXIT_CODE_FAILED;

    public static void main(String[] args) {
        ExtractorTask extractorTask;

        try {
            CliContext cliContext = createCliContext(args);

            if (cliContext != null) {
                extractorTask = getInitializedExtractorTask(cliContext);

                runExtractorTasks(extractorTask);

                exitCode = EXIT_CODE_SUCCESS;

                LOG.info("Successfully completed execution of extraction task.");
            } else {
                exitCode = EXIT_CODE_HELP;
            }
        } catch (Exception e) {
            LOG.error("Error encountered.", e);

            System.out.println(e.getMessage());
        }

        System.exit(exitCode);
    }

    @VisibleForTesting
    static CliContext createCliContext(String[] args) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Initializing the cli context.");
        }

        Options acceptedCliOptions = prepareCommandLineOptions();

        try {
            CommandLine  cmd              = new DefaultParser().parse(acceptedCliOptions, args, true);
            List<String> argsNotProcessed = cmd.getArgList();

            if (argsNotProcessed != null && argsNotProcessed.size() > 0) {
                throw new AtlasBaseException("Unrecognized arguments.");
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Successfully initialized the cli context.");
            }

            CliContext ret = null;

            if (cmd.hasOption(CliContext.OPTION_EXTRACTION_HELP)) {
                printUsage(acceptedCliOptions);
            } else {
                ret = CliContext.createInstance(cmd);
            }

            return ret;
        } catch (ParseException | AtlasBaseException e) {
            printUsage(acceptedCliOptions);

            throw new AtlasBaseException("Invalid arguments. Reason: " + e.getMessage(), e);
        }
    }

    @VisibleForTesting
    static ExtractorTask getInitializedExtractorTask(CliContext cliContext) throws AtlasBaseException {
        ExtractorTask extractorTask;

        try {
            LOG.info("Initializing the extractor task.");

            extractorTask = ExtractorTaskFactory.getExtractorTask(cliContext);

            extractorTask.initialize();

            LOG.info("Successfully initialized the extractor task.");

            return extractorTask;
        } catch (Throwable e) {
            throw new AtlasBaseException("Error encountered in initializing task" + (StringUtils.isNotBlank(e.getMessage()) ? ": " + e.getMessage() : ""), e);
        }
    }

    @VisibleForTesting
    static void runExtractorTasks(ExtractorTask extractorTask) throws AtlasBaseException {
        if (extractorTask != null) {
            LOG.info("Running the extractor task {}.", extractorTask.getClass().getName());

            extractorTask.execute();

            LOG.info("Completed the execution of task: " + extractorTask.getName());
        } else {
            LOG.info("Extractor task is null. Nothing to be done.");
        }
    }

    private static Options prepareCommandLineOptions() {
        Options options = new Options();

        return options.addOption(CliContext.OPTION_EXTRACTION_TYPE_SHORT, CliContext.OPTION_EXTRACTION_TYPE_LONG, true,
                "Extraction type for Azure Data Lake Storage . Possible values are “INCREMENTAL” and “BULK” extractions.")
                .addOption(CliContext.OPTION_EXTRACTION_CONFIGURATION, CliContext.OPTION_EXTRACTION_CONFIGURATION_LONG, true,
                        "Configuration file for the Azure Data Lake Storage extraction")
                .addOption(CliContext.OPTION_EXTRACTION_FAIL_ON_ERROR, CliContext.OPTION_EXTRACTION_FAIL_ON_ERROR_LONG, false,
                        "Fail when error is encountered")
                .addOption(CliContext.OPTION_EXTRACTION_HELP, CliContext.OPTION_EXTRACTION_HELP_LONG, false,
                        "Print this help message");
    }

    private static void printUsage(Options options) {
        HelpFormatter formatter = new HelpFormatter();

        formatter.printHelp(AZURE_EXTRACTOR_SCRIPT_NAME, options);
    }
}