/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.cli;

import org.apache.atlas.exception.AtlasBaseException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.net.URL;

/**
 * This is the class that provides the context representing
 * the CLI command line options/parameters.
 */
public class CliContext {
    private static final Logger LOG = LoggerFactory.getLogger(CliContext.class);

    public enum SourceType { ADLS }

    public enum ExtractionType { BULK, INCREMENTAL }

    static final String OPTION_EXTRACTION_TYPE_SHORT         = "e";
    static final String OPTION_EXTRACTION_TYPE_LONG          = "extraction";
    static final String OPTION_EXTRACTION_CONFIGURATION      = "c";
    static final String OPTION_EXTRACTION_CONFIGURATION_LONG = "config";
    static final String OPTION_EXTRACTION_FAIL_ON_ERROR      = "f";
    static final String OPTION_EXTRACTION_FAIL_ON_ERROR_LONG = "failOnError";
    static final String OPTION_EXTRACTION_HELP               = "h";
    static final String OPTION_EXTRACTION_HELP_LONG          = "help";
    static final String DEFAULT_ADLS_CONFIG_FILE_NAME        = "adls.conf";

    private final boolean        failOnError;
    private final File           configFile;
    private final SourceType     sourceType;
    private final ExtractionType extractionType;

    public CliContext(CommandLine cmd) throws AtlasBaseException {
        this.failOnError    = cmd.hasOption(OPTION_EXTRACTION_FAIL_ON_ERROR);
        this.sourceType     = SourceType.ADLS;
        this.extractionType = getExtractionType(cmd);
        this.configFile     = getConfigFile(cmd);

        LOG.info("Preparing extractor with extraction type {}.", this.extractionType.name());
    }

    public static CliContext createInstance(CommandLine cmd) throws AtlasBaseException {
        return new CliContext(cmd);
    }

    public File getConfigFile() {
        return configFile;
    }

    public boolean isFailOnError() {
        return failOnError;
    }

    public SourceType getSourceType() {
        return sourceType;
    }

    public ExtractionType getExtractionType() {
        return extractionType;
    }

    private File getConfigFile(CommandLine cmd) throws AtlasBaseException {
        String configFileName = cmd.getOptionValue(OPTION_EXTRACTION_CONFIGURATION);

        if (StringUtils.isEmpty(configFileName)) {
            String msg = "Missing extraction configuration file parameter. Using the default file: " + DEFAULT_ADLS_CONFIG_FILE_NAME;

            LOG.info(msg);

            configFileName = DEFAULT_ADLS_CONFIG_FILE_NAME;

            LOG.info("Looking for {} in classpath", configFileName);

            URL url = CliContext.class.getClassLoader().getResource(configFileName);

            if (url == null) {
                LOG.info("Looking for /{} in classpath", configFileName);

                url = CliContext.class.getClassLoader().getResource("/" + configFileName);
            }

            if (url != null) {
                configFileName = url.getPath();
            }
        }

        File file = new File(configFileName);

        if (!file.exists() || !file.isFile() || !file.canRead()) {
            String msg = "The configuration file " + configFileName + " is not a file or it does not exist or it is not readable.";

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }

        return file;
    }

    private ExtractionType getExtractionType(CommandLine cmd) throws AtlasBaseException {
        String extractionType = cmd.getOptionValue(OPTION_EXTRACTION_TYPE_SHORT);

        if (extractionType == null) {
            LOG.info("No extraction type provided. Defaulting to Bulk extraction.");

            return ExtractionType.BULK;
        }

        try {
            return ExtractionType.valueOf(extractionType.toUpperCase());
        } catch (IllegalArgumentException e) {
            String msg = "Received unsupported extraction type " + extractionType + ".";

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }
}