/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.atlas.adls.extractors;

import com.azure.storage.blob.BlobServiceClient;
import com.azure.storage.blob.BlobServiceClientBuilder;
import com.azure.storage.queue.QueueServiceClient;
import com.azure.storage.queue.QueueServiceClientBuilder;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import com.azure.core.http.HttpClient;
import com.azure.core.http.netty.NettyAsyncHttpClientBuilder;

public class ADLSClient implements AzureClient {
    private static final Logger     LOG      = LoggerFactory.getLogger(ADLSClient.class);
    private static final ADLSClient INSTANCE = new ADLSClient();

    private final AtomicBoolean                       isInitialized;
    private final HashMap<String, BlobServiceClient>  blobServiceClients;
    private final HashMap<String, QueueServiceClient> queueServiceClients;
    private       ADLSExtractorContext                extractorContext;

    public static ADLSClient getInstance() {
        return INSTANCE;
    }

    @Override
    public void initialize(ADLSExtractorContext extractorContext) throws AtlasBaseException {
        if (!isInitialized.get()) {
            if (extractorContext != null) {
                this.extractorContext = extractorContext;

                for (String accountName : extractorContext.getAllowableAccountNames()) {
                    createAndPersistBlobServiceClient(accountName);
                    createAndPersistQueueServiceClient(accountName);
                }

                String azureAccountName = extractorContext.getAzureAccountNameConfig();

                if (StringUtils.isNotBlank(azureAccountName)) {
                    if (!blobServiceClients.containsKey(azureAccountName)) {
                        createAndPersistBlobServiceClient(azureAccountName);
                    }
                    if (!queueServiceClients.containsKey(azureAccountName)) {
                        createAndPersistQueueServiceClient(azureAccountName);
                    }
                }

                isInitialized.set(true);
            } else {
                throw new AtlasBaseException("Invalid extractor context");
            }
        }
    }

    @Override
    public void reinitialize() throws AtlasBaseException {
        LOG.info("Reconnecting to Azure Data Lake Storage");

        isInitialized.set(false);

        initialize(extractorContext);
    }

    public BlobServiceClient getBlobServiceClient(String accountName) throws AtlasBaseException {
        if (isInitialized.get()) {
            return this.blobServiceClients.get(accountName);
        }

        throw new AtlasBaseException("Azure Data Lake Storage client is not initialized");
    }

    public QueueServiceClient getQueueServiceClient(String accountName) throws AtlasBaseException {
        if (isInitialized.get()) {
            return this.queueServiceClients.get(accountName);
        }

        throw new AtlasBaseException("Azure Data Lake Storage client is not initialized");
    }

    private ADLSClient() {
        this.isInitialized       = new AtomicBoolean(false);
        this.blobServiceClients  = new HashMap<>();
        this.queueServiceClients = new HashMap<>();
    }

    private void createAndPersistBlobServiceClient(String accountName) {
        String                   endpoint                 = String.format(Constants.BLOB_SERVICE_URL_ENDPOINT_FORMAT, accountName);
        HttpClient               client                   = new NettyAsyncHttpClientBuilder().build();
        BlobServiceClientBuilder blobServiceClientBuilder = new BlobServiceClientBuilder().endpoint(endpoint).httpClient(client);

        new ADLSCredentialsProvider(extractorContext).setCredential(blobServiceClientBuilder);

        this.blobServiceClients.put(accountName, blobServiceClientBuilder.buildClient());
    }

    private void createAndPersistQueueServiceClient(String accountName) {
        String                    endpoint                  = String.format(Constants.QUEUE_SERVICE_URL_ENDPOINT_FORMAT, accountName);
        HttpClient                client                    = new NettyAsyncHttpClientBuilder().build();
        QueueServiceClientBuilder queueServiceClientBuilder = new QueueServiceClientBuilder().endpoint(endpoint).httpClient(client);

        new ADLSCredentialsProvider(extractorContext).setCredential(queueServiceClientBuilder);

        this.queueServiceClients.put(accountName, queueServiceClientBuilder.buildClient());
    }
}
