/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.extractors;

import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.BlobServiceClient;
import com.azure.storage.blob.models.BlobContainerItem;
import org.apache.atlas.adls.bridge.ADLSHook;
import org.apache.atlas.adls.bridge.AtlasADLSCreateEvent;
import org.apache.atlas.adls.bridge.models.AtlasADLSAccountEntity;
import org.apache.atlas.adls.bridge.models.AtlasADLSContainerEntity;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.model.instance.AtlasEntity;
import org.apache.atlas.model.instance.AtlasObjectId;
import org.apache.atlas.model.instance.AtlasRelatedObjectId;
import org.apache.atlas.type.AtlasTypeUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;

/**
 * This class handles the Azure Data Lake Storage Container specific extractions.
 */
public class ADLSContainerExtractor implements Serializable {
    private static final long   serialVersionUID              = 1L;
    private static final Logger LOG                           = LoggerFactory.getLogger(ADLSContainerExtractor.class);
    private static final String ADLS_GEN_2_ACCOUNT_CONTAINERS = "adls_gen2_account_containers";
    private static final String ATTRIBUTE_ACCOUNT             = "account";

    private final ADLSExtractorContext extractorContext;

    public ADLSContainerExtractor(ADLSExtractorContext extractorContext) {
        this.extractorContext = extractorContext;
    }

    public Set<String> getAllContainerNames(String accountName) throws AtlasBaseException {
        int               retry             = 0;
        BlobServiceClient blobServiceClient = ADLSClient.getInstance().getBlobServiceClient(accountName);

        if (blobServiceClient != null) {
            while (retry < extractorContext.getMaxReconnectCount()) {
                try {
                    Set<String> ret = new HashSet<>();

                    for (BlobContainerItem eachContainerItem : blobServiceClient.listBlobContainers()) {
                        ret.add(eachContainerItem.getName());
                    }

                    return ret;
                } catch (Exception e) {
                    //Still unable to find more information about the retryable exception. So for now retry for all exceptions
                    retry++;
                }
            }
        }

        throw new AtlasBaseException("Unable to list all the containers for the account Name: " + accountName);
    }

    public BlobContainerClient getContainerClient(String accountName, String containerName) throws AtlasBaseException {
        BlobContainerClient ret     = null;
        int                 retry   = 0;

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                BlobServiceClient blobServiceClient = ADLSClient.getInstance().getBlobServiceClient(accountName);

                if (blobServiceClient != null) {
                    ret = blobServiceClient.getBlobContainerClient(containerName);
                }

                if (ret == null) {
                    break;
                }

                return ret;
            } catch (Exception e) {
                //Still unable to find more information about the retryable exception. So for now retry for all exceptions
                retry++;
            }
        }

        throw new AtlasBaseException("Unable get container client for container: " + containerName);
    }

    public AtlasADLSAccountEntity extractAccount(BlobContainerClient blobContainerClient) throws AtlasBaseException {
        int    retry   = 0;
        String errMsg  = String.format("Azure Data Lake Storage exception occurred while extracting container %s", blobContainerClient.getBlobContainerName());

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                AtlasADLSAccountEntity accountEntity = null;

                if (extractorContext.isContainerAllowListed(getContainerAndAccountNameForLookup(blobContainerClient))) {
                    accountEntity = new AtlasADLSAccountEntity(extractorContext, blobContainerClient);

                    publishCreateAccountToAtlas(accountEntity);
                }

                return accountEntity;
            } catch (Exception e) {
                //Still unable to find more information about the retryable exception. So for now retry for all exceptions
                retry++;
            }
        }

        throw new AtlasBaseException(errMsg);
    }

    public AtlasADLSContainerEntity extractContainer(AtlasADLSAccountEntity accountEntity, BlobContainerClient blobContainerClient) throws AtlasBaseException {
        int                      retry           = 0;
        String                   errMsg          = String.format("Azure Data Lake Storage exception occurred while extracting container %s", blobContainerClient.getBlobContainerName());
        AtlasADLSContainerEntity containerEntity = null;

        while (retry < extractorContext.getMaxReconnectCount()) {
            try {
                if (extractorContext.isContainerAllowListed(getContainerAndAccountNameForLookup(blobContainerClient))) {
                    containerEntity = new AtlasADLSContainerEntity(extractorContext, blobContainerClient);

                    linkAtlasADLSAccountToContainer(accountEntity.getEntity(), containerEntity.getEntity());

                    publishCreateContainerToAtlas(containerEntity);
                }

                return containerEntity;
            } catch (Exception e) {
                //Still unable to find more information about the retryable exception. So for now retry for all exceptions
                retry++;
            }
        }

        throw new AtlasBaseException(errMsg);
    }

    public void publishCreateAccountToAtlas(AtlasADLSAccountEntity accountEntity) throws AtlasBaseException {
        if (accountEntity != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Publish create Account entity event to Atlas.");
            }

            try {
                AtlasADLSCreateEvent azureADLSCreateEvent = new AtlasADLSCreateEvent(accountEntity.getAtlasEntitiesWithExtInfo());

                ADLSHook.getInstance().publishEvent(azureADLSCreateEvent);
            } catch (Exception e) {
                String msg = String.format("Error encountered while publishing create account entity with Account Name {%s}." +
                        "Extraction process will continue with remaining containers.", accountEntity.name());

                LOG.error(msg, e);

                throw new AtlasBaseException(msg, e);
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Published Azure Data Lake Storage Account {} entity create event to Atlas.", accountEntity.name());
            }
        }
    }

    public void publishCreateContainerToAtlas(AtlasADLSContainerEntity containerEntity) throws AtlasBaseException {
        if (containerEntity != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Publish create container entity event to Atlas.");
            }

            try {
                AtlasADLSCreateEvent azureADLSCreateEvent = new AtlasADLSCreateEvent(containerEntity.getAtlasEntitiesWithExtInfo());

                ADLSHook.getInstance().publishEvent(azureADLSCreateEvent);
            } catch (Exception e) {
                String msg = String.format("Error encountered while publishing create container entity for container{%s}." +
                        "Extraction process will continue with remaining containers.", containerEntity.name());

                LOG.error(msg, e);

                throw new AtlasBaseException(msg, e);
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("Published Azure Data Lake Storage Container {} entity create event to Atlas.", containerEntity.name());
            }
        }
    }

    private void linkAtlasADLSAccountToContainer(AtlasEntity accountEntity, AtlasEntity containerAtlasEntity) {
        accountEntity.setGuid(null);

        AtlasObjectId        containerEntityId = AtlasTypeUtil.getObjectId(accountEntity);
        AtlasRelatedObjectId relatedObjectId   = new AtlasRelatedObjectId(containerEntityId, ADLS_GEN_2_ACCOUNT_CONTAINERS);

        containerAtlasEntity.setRelationshipAttribute(ATTRIBUTE_ACCOUNT, relatedObjectId);
    }

    String getContainerAndAccountNameForLookup(BlobContainerClient blobContainerClient) {
        return blobContainerClient.getBlobContainerName() + Constants.DEFAULT_CONTAINER_ACCOUNT_DELIMITER + blobContainerClient.getAccountName();
    }
}