/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.extractors.incremental;

import org.apache.atlas.adls.extractors.Constants;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.extractors.ADLSExtractorContext;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.CollectionUtils;

import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

/**
 * A context class that contains shared information for extraction
 */
public class ADLSIncrementalExtractorContext extends ADLSExtractorContext {
    private static final long serialVersionUID = 1L;

    private static final Logger LOG = LoggerFactory.getLogger(ADLSIncrementalExtractorContext.class);

    private static final int TOTAL_FIELDS_PER_ACCOUNT_QUEUE_CONFIG = 2;

    public ADLSIncrementalExtractorContext(Properties extractionConfiguration) throws AtlasBaseException {
        super(extractionConfiguration);
    }

    public void validateConfiguration() throws AtlasBaseException {
        super.validateConfiguration();

        checkEmptinessForConfigurationParameter(Constants.CONFIG_QUEUE_NAMES, getQueueName());

        validateMaxNumberOfMsgPerRequestConfig();

        validateSqsRequestWaitTimeConfig();

        validateMaxRetryConfig();

        validateVisibilityTimeoutConfig();
    }

    public void initialize() throws AtlasBaseException {
        super.initialize();

        loadQueueNames();
    }

    public Integer getMaxNumberOfMessagesPerRequest() {
        String maxMsgsPerRequest = extractionConfiguration.getProperty(Constants.CONFIG_QUEUE_MAX_MESSAGES_PER_REQUEST);

        if (maxMsgsPerRequest == null) {
            return Constants.DEFAULT_QUEUE_MAX_MESSAGES_PER_REQUEST;
        } else {
            return Integer.parseInt(maxMsgsPerRequest);
        }
    }

    public Integer getVisibilityTimeout() {
        String visibilityTimeout = extractionConfiguration.getProperty(Constants.CONFIG_QUEUE_VISIBILITY_TIMEOUT);

        if (visibilityTimeout == null) {
            return Constants.DEFAULT_QUEUE_VISIBILITY_TIMEOUT;
        } else {
            return Integer.parseInt(visibilityTimeout);
        }
    }

    public Integer getWaitTimeInSeconds() {
        String waitTime = extractionConfiguration.getProperty(Constants.CONFIG_QUEUE_REQUEST_WAIT_TIME);

        if (waitTime == null) {
            return Constants.DEFAULT_QUEUE_REQUEST_WAIT_TIME;
        } else {
            return Integer.parseInt(waitTime);
        }
    }

    public int getMaxRetryCount() {
        String maxRetryCount = extractionConfiguration.getProperty(Constants.CONFIG_MAX_RETRY);

        if (StringUtils.isEmpty(maxRetryCount)) {
            return Constants.DEFAULT_MAX_RETRY;
        } else {
            return Integer.parseInt(maxRetryCount);
        }
    }

    private void validateMaxNumberOfMsgPerRequestConfig() throws AtlasBaseException {
        String msg = generateInvalidConfigurationMessage(Constants.CONFIG_QUEUE_MAX_MESSAGES_PER_REQUEST);

        try {
            Integer maxMsgPerRequest = getMaxNumberOfMessagesPerRequest();

            if (maxMsgPerRequest < Constants.MIN_FOR_QUEUE_MAX_MESSAGES_PER_REQUEST
                    || maxMsgPerRequest > Constants.MAX_FOR_QUEUE_MAX_MESSAGES_PER_REQUEST) {
                LOG.error(msg);

                throw new AtlasBaseException(msg);
            }
        } catch (NumberFormatException e) {
            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    private void validateVisibilityTimeoutConfig() throws AtlasBaseException {
        try {
            getVisibilityTimeout();
        } catch (NumberFormatException e) {
            String msg = generateInvalidConfigurationMessage(Constants.CONFIG_QUEUE_VISIBILITY_TIMEOUT);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    private void validateSqsRequestWaitTimeConfig() throws AtlasBaseException {
        try {
            getWaitTimeInSeconds();
        } catch (NumberFormatException e) {
            String msg = generateInvalidConfigurationMessage(Constants.CONFIG_QUEUE_REQUEST_WAIT_TIME);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    private void validateMaxRetryConfig() throws AtlasBaseException {
        try {
            getMaxRetryCount();
        } catch (NumberFormatException e) {
            String msg = generateInvalidConfigurationMessage(Constants.CONFIG_MAX_RETRY);

            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }
    }

    private void loadQueueNames() throws AtlasBaseException {
        for (String eachAccountQueue : StringUtils.split(extractionConfiguration.getProperty(Constants.CONFIG_QUEUE_NAMES), Constants.CONFIG_PATH_SEPARATOR)) {
            if (StringUtils.isNotBlank(eachAccountQueue)) {
                String[] accountNameQueue = StringUtils.split(StringUtils.trim(eachAccountQueue), Constants.CONFIG_ACCOUNT_QUEUE_SEPARATOR);

                if (ArrayUtils.isNotEmpty(accountNameQueue) && accountNameQueue.length == TOTAL_FIELDS_PER_ACCOUNT_QUEUE_CONFIG) {
                    String accountName = accountNameQueue[0];
                    String queueName   = accountNameQueue[1];

                    if (accountAndQueueNames.containsKey(accountName)) {
                        accountAndQueueNames.get(accountName).add(queueName);
                    } else {
                        Set<String> queueNameSet = new TreeSet<>();

                        queueNameSet.add(queueName);

                        accountAndQueueNames.put(accountName, queueNameSet);
                    }

                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Adding account:" + accountName + " and queue name:" + queueName);
                    }
                }
            }
        }

        validateQueueNameConfig();
    }

    private void validateQueueNameConfig() throws AtlasBaseException {
        String msg = generateInvalidConfigurationMessage(Constants.CONFIG_QUEUE_NAMES);

        if (CollectionUtils.isEmpty(accountAndQueueNames)) {
            LOG.error(msg);

            throw new AtlasBaseException(msg);
        }

        Set<String> accountNames = new HashSet<>();

        if (isAllowListingConfigurationPresent()) {
            accountNames = getAllowableAccountNames();
        }

        if (!accountNames.contains(getAzureAccountNameConfig())) {
            accountNames.add(getAzureAccountNameConfig());
        }

        for (String accountName: accountNames) {
            if (!accountAndQueueNames.containsKey(accountName)) {
                msg += ". Details: " + "Missing queue configuration for Account Name: " + accountName;

                LOG.error(msg);

                throw new AtlasBaseException(msg);
            }
        }
    }
}