/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.atlas.adls.extractors.incremental;

import com.azure.core.http.rest.PagedResponse;
import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.models.BlobItem;
import com.azure.storage.queue.QueueClient;
import com.azure.storage.queue.models.QueueMessageItem;
import com.microsoft.azure.eventgrid.customization.EventGridSubscriber;
import com.microsoft.azure.eventgrid.models.EventGridEvent;

import org.apache.atlas.adls.bridge.models.*;
import org.apache.atlas.adls.extractors.ADLSBlobExtractor;
import org.apache.atlas.adls.extractors.ADLSClient;
import org.apache.atlas.adls.extractors.ADLSContainerExtractor;
import org.apache.atlas.exception.AtlasBaseException;
import org.apache.atlas.adls.extractors.Constants;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.Serializable;
import java.time.Duration;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

public class QueueMsgHandler implements Serializable{
    private static final long serialVersionUID = 1L;
    private static final Logger LOG = LoggerFactory.getLogger(QueueMsgHandler.class);

    private final ADLSIncrementalExtractorContext       adlsIncrementalExtractorContext;
    private final ADLSContainerExtractor                adlsContainerExtractor;
    private final ADLSBlobExtractor                     adlsBlobExtractor;

    private enum ADLSEventType {
        MICROSOFT_STORAGE_BLOBCREATED("Microsoft.Storage.BlobCreated"),
        MICROSOFT_STORAGE_BLOBDELETED("Microsoft.Storage.BlobDeleted"),
        MICROSOFT_STORAGE_BLOBRENAMED("Microsoft.Storage.BlobRenamed"),
        MICROSOFT_STORAGE_DIRECTORYCREATED("Microsoft.Storage.DirectoryCreated"),
        MICROSOFT_STORAGE_DIRECTORYDELETED("Microsoft.Storage.DirectoryDeleted"),
        MICROSOFT_STORAGE_DIRECTORYRENAMED("Microsoft.Storage.DirectoryRenamed");

        private final String name;

        ADLSEventType(String name) {
            this.name = name;
        }

        public String getName() {
            return name.toUpperCase();
        }

        public static ADLSEventType getEventType(EventGridEvent event) {
            return ADLSEventType.valueOf(event.eventType().replace('.', '_').toUpperCase());
        }

        public static boolean isDirectoryEvent(EventGridEvent event) {
            ADLSEventType eventType = getEventType(event);

            return (eventType == MICROSOFT_STORAGE_DIRECTORYCREATED || eventType == MICROSOFT_STORAGE_DIRECTORYDELETED || eventType == MICROSOFT_STORAGE_DIRECTORYRENAMED);
        }
    }

    public QueueMsgHandler(ADLSIncrementalExtractorContext adlsIncrementalExtractorContext) {
        this.adlsIncrementalExtractorContext = adlsIncrementalExtractorContext;
        this.adlsContainerExtractor          = new ADLSContainerExtractor(adlsIncrementalExtractorContext);
        this.adlsBlobExtractor               = new ADLSBlobExtractor(adlsIncrementalExtractorContext);
    }

    public void initialize() throws AtlasBaseException {
        ADLSClient.getInstance().initialize(adlsIncrementalExtractorContext);

        loadADLSContainersIntoCache();
    }

    public int processQueueMessages(String account, String queueName) throws AtlasBaseException {
        int         retTotalProcessedMessages = 0;
        QueueClient queueClient               = ADLSClient.getInstance().getQueueServiceClient(account).getQueueClient(queueName);

        LOG.info("Read messages from queue {} for account{}.", queueClient.getQueueName(), queueClient.getAccountName());

        // Received all messages in queue
        for (PagedResponse<QueueMessageItem> resp : queueClient.receiveMessages(adlsIncrementalExtractorContext.getMaxNumberOfMessagesPerRequest(),
                Duration.ofSeconds(adlsIncrementalExtractorContext.getVisibilityTimeout()),
                Duration.ofSeconds(adlsIncrementalExtractorContext.getWaitTimeInSeconds()), null).iterableByPage()) {

            for (QueueMessageItem queueMessage : resp.getValue()) {
                try {
                    if (StringUtils.isNotBlank(queueMessage.getMessageText())) {
                        String           decodedMessageText = new String(Base64.decodeBase64(queueMessage.getMessageText().getBytes()));
                        EventGridEvent[] eventGridEvents    = new EventGridSubscriber().deserializeEventGridEvents(decodedMessageText);

                        for (EventGridEvent eventRecord : eventGridEvents) {
                            processEventRecord(account, eventRecord);

                            queueClient.deleteMessage(queueMessage.getMessageId(), queueMessage.getPopReceipt());

                            retTotalProcessedMessages++;
                        }
                    }
                } catch (AtlasBaseException | IOException e) {
                    String msg = String.format("Error encountered while processing Azure Data Lake Storage Queue message. Reason: %s", e.getMessage());

                    LOG.warn(msg);

                    if (adlsIncrementalExtractorContext.getFailOnErrorFlag()) {
                        throw new AtlasBaseException(msg, e);
                    }
                }
            }
        }

        return retTotalProcessedMessages;
    }

    private boolean isValidEvent(EventGridEvent eventRecord) {
        return isEventRecordStructureValid(eventRecord) &&
               isAtlasSupportedEventType(eventRecord) &&
               isEventOccurredAfterBulkExtractionStart(eventRecord.eventTime().getMillis());
    }

    private void loadADLSContainersIntoCache() throws AtlasBaseException {
        AtlasADLSContainerCache.getInstance().getAtlasADLSContainerCache().clear();

        Map<String, Set<String>> accountContainerList;

        if (adlsIncrementalExtractorContext.isAllowListingConfigurationPresent()) {
            accountContainerList = adlsIncrementalExtractorContext.getAllowableAccountAndContainers();
        } else {
            accountContainerList = new HashMap<>();

            String      accountName                 = adlsIncrementalExtractorContext.getAzureAccountNameConfig();
            Set<String> allContainerNamesForAccount = adlsContainerExtractor.getAllContainerNames(accountName);

            accountContainerList.put(accountName, allContainerNamesForAccount);
        }

        for (Map.Entry<String, Set<String>> eachEntry : accountContainerList.entrySet()) {
            AtlasADLSAccountEntity accountEntity = null;

            for (String eachContainer : eachEntry.getValue()) {
                String              accountName         = eachEntry.getKey();
                BlobContainerClient blobContainerClient = adlsContainerExtractor.getContainerClient(accountName, eachContainer);

                try {
                    if (accountEntity == null) {
                        accountEntity = adlsContainerExtractor.extractAccount(blobContainerClient);
                    }

                    AtlasADLSContainerEntity azureADLSContainerEntity = adlsContainerExtractor.extractContainer(accountEntity, blobContainerClient);

                    if (azureADLSContainerEntity != null) {
                        AtlasADLSContainerCache.getInstance().getAtlasADLSContainerCache().put(formKeyForCache(accountName, blobContainerClient.getBlobContainerName()), azureADLSContainerEntity);
                    }
                } catch (AtlasBaseException e) {
                    LOG.info("Unable to load container{} to cache for account{}.", eachContainer, accountName);

                    if (adlsIncrementalExtractorContext.getFailOnErrorFlag()) {
                        throw new AtlasBaseException(e.getMessage(), e);
                    }
                }
            }
        }
    }

    private String formKeyForCache(String accountName, String container) {
        return accountName + adlsIncrementalExtractorContext.getPrefixDelimiter() + container;
    }

    private boolean isEventRecordStructureValid(EventGridEvent gridEvent) {
        boolean ret = false;

        if (gridEvent.eventType() != null && gridEvent.data() != null && StringUtils.isNotEmpty(gridEvent.subject()) && gridEvent.eventTime() != null) {
            ret = true;
        } else {
            LOG.info("Skipping unsupported event record structure with event name {} and event time {}",
                    gridEvent.eventType(), gridEvent.eventTime());
        }

        return ret;
    }

    private boolean isAtlasSupportedEventType(EventGridEvent event) {
        boolean ret = false;

        try {
            ADLSEventType.getEventType(event);

            ret = true;
        } catch (IllegalArgumentException e) {
            LOG.info("Skipping unsupported event type {}.", event.eventType());
        }

        return ret;
    }

    private boolean isEventOccurredAfterBulkExtractionStart(long eventTime) {
        boolean isEventOccurredAfterBulk = true;
        String  lastExtractionTime       = adlsIncrementalExtractorContext.getPropertyFromProgressFile(Constants.PROPERTY_EXTRACTION_START_TIME_AT_PROGRESS_FILE);

        if (StringUtils.isNotEmpty(lastExtractionTime)) {
            if (eventTime < Long.parseLong(lastExtractionTime)) {
                isEventOccurredAfterBulk = false;

                LOG.info("Skipping event with event time {} which has been reported before last successful bulk extraction start time {}.",
                        eventTime, lastExtractionTime);
            }
        }

        return isEventOccurredAfterBulk;
    }

    public void processEventRecord(String accountName, EventGridEvent gridEvent) throws AtlasBaseException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Processing event {} for blob/directory {} ", gridEvent.eventType(), gridEvent.data());
        }

        if (isValidEvent(gridEvent)) {
            String containerName = getContainerNameFromEventSubject(gridEvent.subject());
            AtlasADLSContainerEntity containerEntityFromCache = getAtlasADLSContainerEntityFromCache(accountName, containerName);
            BlobContainerClient blobContainerClient = adlsContainerExtractor.getContainerClient(accountName, containerName);

            if (containerEntityFromCache != null) {
                String blobName = getBlobNameFromEventSubject(gridEvent.subject());
                boolean isDirectory = ADLSEventType.isDirectoryEvent(gridEvent);
                BlobItem blobItem = adlsBlobExtractor.getBlobWithMoreDetailsIfPresent(blobContainerClient, blobName, isDirectory);
                BlobItem oldBlob;

                switch (ADLSEventType.getEventType(gridEvent)) {
                    case MICROSOFT_STORAGE_BLOBCREATED:
                    case MICROSOFT_STORAGE_DIRECTORYCREATED:
                        adlsBlobExtractor.createADLSDirOrBlobEntitiesAtAtlas(blobItem, containerEntityFromCache);
                        break;

                    case MICROSOFT_STORAGE_BLOBDELETED:
                    case MICROSOFT_STORAGE_DIRECTORYDELETED:
                        adlsBlobExtractor.deleteEntityAtAtlas(blobItem, containerEntityFromCache);
                        break;

                    case MICROSOFT_STORAGE_BLOBRENAMED:
                        oldBlob = createADLSOldBlobItemForRename(gridEvent);
                        adlsBlobExtractor.updateEntityAtAtlas(oldBlob, blobItem, containerEntityFromCache, containerEntityFromCache);
                        break;

                    case MICROSOFT_STORAGE_DIRECTORYRENAMED:
                        oldBlob = createADLSOldBlobItemForRename(gridEvent);
                        BlobContainerClient containerClient = adlsContainerExtractor.getContainerClient(accountName, containerName);
                        adlsBlobExtractor.updateEntireDirectoryContentAtAtlas(containerClient, oldBlob, blobItem, containerEntityFromCache);
                        break;
                }
            } else {
                LOG.info("Skipping the event with account {}, container {} and blob{} generated.",
                        accountName,
                        getContainerNameFromEventSubject(gridEvent.subject()),
                        getBlobNameFromEventSubject(gridEvent.subject()));
            }
        }
    }

    private AtlasADLSContainerEntity getAtlasADLSContainerEntityFromCache(String accountName, String container) {
        String containerAndAccountNameForCacheLookup = formKeyForCache(accountName, container);

        return AtlasADLSContainerCache.getInstance().getAtlasADLSContainerCache().get(containerAndAccountNameForCacheLookup);
    }

    private BlobItem createADLSOldBlobItemForRename(EventGridEvent record) throws AtlasBaseException {
        BlobItem sourceBlobItem = new BlobItem();

        if (ADLSEventType.isDirectoryEvent(record)) {
            Map<String, String> metadata = new HashMap<>();

            metadata.put(Constants.IS_DIRECTORY_METADATA_KEY, Boolean.TRUE.toString());

            sourceBlobItem.setMetadata(metadata);
        }

        if (record.data() != null && record.data() instanceof Map) {
            Map<String, String> eventData = (Map<String, String>) record.data();

            sourceBlobItem.setName(getBlobNameFromUrl(eventData.get(Constants.EVENT_DATA_SOURCE_URL)));
        }

        return sourceBlobItem;
    }

    private String getContainerNameFromEventSubject(String subject) {
        String ret = "";

        if (StringUtils.isNotBlank(subject) && subject.indexOf(Constants.DEFAULT_SUBJECT_PREFIX) == 0) {
            ret                      = Constants.DEFAULT_SUBJECT_PREFIX.length() < subject.length() ? subject.substring(Constants.DEFAULT_SUBJECT_PREFIX.length()) : "";
            int containerEndPosition = ret.indexOf(adlsIncrementalExtractorContext.getPrefixDelimiter());

            if (containerEndPosition > -1) {
                ret = ret.substring(0, containerEndPosition);
            }
        }
        return ret;
    }

    private String getBlobNameFromEventSubject(String subject) {
        String ret = "";

        if (StringUtils.isNotBlank(subject) && subject.contains(Constants.DEFAULT_SUBJECT_BLOB_NAME_SEPARATOR)) {
            int startPositionForBlobName = subject.indexOf(Constants.DEFAULT_SUBJECT_BLOB_NAME_SEPARATOR) + Constants.DEFAULT_SUBJECT_BLOB_NAME_SEPARATOR.length();

            ret = startPositionForBlobName < subject.length() ? subject.substring(startPositionForBlobName) : "";
        }

        return ret;
    }

    private String getBlobNameFromUrl(String url) throws AtlasBaseException {
        String blobName = "";

        if (StringUtils.isNotBlank(url) && url.contains(Constants.DEFAULT_ADLS_GEN2_ACCOUNT_HOST_SUFFIX)) {
            int startPositionForContainer = url.indexOf(Constants.DEFAULT_ADLS_GEN2_ACCOUNT_HOST_SUFFIX) + Constants.DEFAULT_ADLS_GEN2_ACCOUNT_HOST_SUFFIX.length() + 1;

            if (startPositionForContainer < url.length()) {
                String urlWithContainerAndBlobPrefix = url.substring(startPositionForContainer);
                int    startPosForBlobPrefix         = urlWithContainerAndBlobPrefix.indexOf(adlsIncrementalExtractorContext.getPrefixDelimiter()) + 1;

                if (startPosForBlobPrefix < urlWithContainerAndBlobPrefix.length()) {
                    blobName = urlWithContainerAndBlobPrefix.substring(startPosForBlobPrefix);
                }
            }
        } else {
            throw new AtlasBaseException("Invalid Blob url: " + url);
        }

        return blobName;
    }
}